/*
 * nodewatcher - collect system information of routers running openwrt
 *               and print or store them into a file
 * Copyright (C) 2013 Clemens John <clemens@freifunk-ol.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/* 
 * Informations about how to start developing with uci in C can be found here:
 * http://wiki.openwrt.org/doc/techref/uci#usage.outside.of.openwrt
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <uci.h>

/* for hostname */
#include <unistd.h>

/* for uptime */
#include <linux/kernel.h>       /* for struct sysinfo */

/* for kernel version */
#include <sys/utsname.h>

/* for interfaces */
#include <sys/ioctl.h>
#include <arpa/inet.h>

/* ip test */
//#include <utils.h>
#include <libnetlink.h>

#include <linux/if_arp.h>
#include <fnmatch.h>

/* for current time */
#include <time.h>
/* for sysinfo */
#include <sys/sysinfo.h>

/*#include "rt_names.h"*/

typedef struct
{
	__u8 family;
	__u8 bytelen;
	__s16 bitlen;
	__u32 flags;
	__u32 data[8];
} inet_prefix;

#define SPRINT_BSIZE 64
#define SPRINT_BUF(x)	char x[SPRINT_BSIZE]

#ifndef nodewatcher_h
#define nodewatcher_h

#define MAXINTERFACES 20
#define SIZEOFMAC 18

extern int ll_remember_index(const struct sockaddr_nl *who,
			     struct nlmsghdr *n, void *arg);
extern int ll_init_map(struct rtnl_handle *rth);
extern unsigned ll_name_to_index(const char *name);
extern const char *ll_index_to_name(unsigned idx);
extern const char *ll_idx_n2a(unsigned idx, char *buf);
extern int ll_index_to_type(unsigned idx);
extern unsigned ll_index_to_flags(unsigned idx);
extern unsigned ll_index_to_addr(unsigned idx, unsigned char *addr,
				 unsigned alen);

/**
 * @brief Struct to store informations read from /proc/cpuinfo
 */
struct cpuinfo {
	char cpu[100];
	char chipset[100];
};

/**
 * @brief Struct to store informations read from /etc/openwrt_release
 */
struct openwrtinfo {
	char distrib_id[100];
	char distrib_release[100];
};

/**
 * @brief Struct to store informations read from /etc/firmware_release
 */
struct firmwareinfo {
	char firmware_version[100];
	char firmware_revision[100];
	char openwrt_core_revision[100];
	char openwrt_feeds_packages_revision[100];
};

/**
 * @brief method to parse /etc/firmware_release
 * @param s_openwrtinfo pointer to a struct where the parsed data will be stored
 * @return 0 if everything was okay, 1 otherwise
 */
int get_firmwareinfo(struct firmwareinfo *s_firmwareinfo)
{
	FILE* fp; 
	char buffer[4096]; 
	size_t bytes_read;
	char* match;
	
	/* Read the entire contents of /proc/cpuinfo into the buffer.  */ 
	fp = fopen ("/etc/firmware_release", "r"); 
	bytes_read = fread (buffer, 1, sizeof (buffer), fp); 
	fclose (fp); 
	
	/* Bail if read failed or if buffer isn't big enough.  */
	if (bytes_read == 0 || bytes_read == sizeof (buffer)) 
		return 1; 
	
	/* Locate the line  */ 
	match = strstr (buffer, "FIRMWARE_VERSION"); 
	if (match != NULL)
	{
		/* Parse the line 
		 * http://docs.roxen.com/pike/7.0/tutorial/strings/sscanf.xml */
		sscanf(match, "FIRMWARE_VERSION=\"%[a-zA-Z .:0-9]\"", s_firmwareinfo->firmware_version);
	}
	
	/* Locate the line  */ 
	match = strstr (buffer, "FIRMWARE_REVISION"); 
	if (match != NULL)
	{
		/* Parse the line 
		   http://docs.roxen.com/pike/7.0/tutorial/strings/sscanf.xml */
		sscanf(match, "FIRMWARE_REVISION=\"%[a-zA-Z .:0-9]\"", s_firmwareinfo->firmware_revision);
	}
	
	/* Locate the line  */ 
	match = strstr (buffer, "OPENWRT_CORE_REVISION"); 
	if (match != NULL)
	{
		/* Parse the line 
		   http://docs.roxen.com/pike/7.0/tutorial/strings/sscanf.xml */
		sscanf(match, "OPENWRT_CORE_REVISION=\"%[a-zA-Z 1-9]\"", s_firmwareinfo->openwrt_core_revision);
	}
	
	/* Locate the line  */ 
	match = strstr (buffer, "OPENWRT_FEEDS_PACKAGES_REVISION"); 
	if (match != NULL)
	{
		/* Parse the line 
		   http://docs.roxen.com/pike/7.0/tutorial/strings/sscanf.xml */
		sscanf(match, "OPENWRT_FEEDS_PACKAGES_REVISION=\"%[a-zA-Z 1-9]\"", s_firmwareinfo->openwrt_feeds_packages_revision);
	}
	return 0;
}

/**
 * @brief method to parse /etc/openwrt_release
 * @param s_openwrtinfo pointer to a struct where the parsed data will be stored
 * @return 0 if everything was okay, 1 otherwise
 */
int get_openwrtinfo(struct openwrtinfo *s_openwrtinfo)
{
	FILE* fp; 
	char buffer[4096]; 
	size_t bytes_read;
	char* match;
	
	/* Read the entire contents of /proc/cpuinfo into the buffer.  */ 
	fp = fopen ("/etc/openwrt_release", "r"); 
	bytes_read = fread (buffer, 1, sizeof (buffer), fp); 
	fclose (fp); 
	
	/* Bail if read failed or if buffer isn't big enough.  */
	if (bytes_read == 0 || bytes_read == sizeof (buffer)) 
		return 1; 
	
	/* Locate the line  */ 
	match = strstr (buffer, "DISTRIB_RELEASE"); 
	if (match != NULL)
	{
		/* Parse the line 
		 * http://docs.roxen.com/pike/7.0/tutorial/strings/sscanf.xml */
		sscanf(match, "DISTRIB_RELEASE=\"%[a-zA-Z 1-9]\"", s_openwrtinfo->distrib_release);
	}
	
	/* Locate the line  */ 
	match = strstr (buffer, "DISTRIB_ID"); 
	if (match != NULL)
	{
		/* Parse the line  */
		sscanf(match, "DISTRIB_ID=\"%[a-zA-Z 1-9]\"", s_openwrtinfo->distrib_id);
	}
	return 0;
}

/**
 * @brief method to parse /proc/cpuinfo
 * @param s_cpuinfo pointer to a struct where the parsed data will be stored
 * @return 0 if everything was okay, 1 otherwise
 */
int get_cpuinfo(struct cpuinfo *s_cpuinfo)
{
	FILE* fp; 
	char buffer[4096]; 
	size_t bytes_read;
	char* match;
	
	/* Read the entire contents of /proc/cpuinfo into the buffer.  */ 
	fp = fopen ("/proc/cpuinfo", "r"); 
	bytes_read = fread (buffer, 1, sizeof (buffer), fp); 
	fclose (fp); 
	
	/* Bail if read failed or if buffer isn't big enough.  */
	if (bytes_read == 0 || bytes_read == sizeof (buffer)) 
		return 1; 
	
	/* Locate the line  */ 
	match = strstr (buffer, "model"); 
	if (match != NULL)
	{
		/* Parse the line  */
		sscanf(match, "model  :  %s", s_cpuinfo->cpu);
	}
	
	match = strstr (buffer, "system type");
	if (match != NULL)
	{
		/* Parse the line  */
		sscanf(match, "system type  :  %s", s_cpuinfo->chipset);
	}
	
	return 0; 
}

/**
 * @brief returns the value of an option stored in a uci formatted
 *        configuration file. These files have to be stored in /etc/config/
 * @param file char array containing the name of the configuration file
 *             we want to parse. I.e. "nodewatcher"
 * @param config_string contains the string we use to locate the option value
 *                      in the config file. I.e. "@network[0].mesh_interface"
 * @return pointer to a char array containing the value of the option.
 *         The memory allocated for this array needs to be freed in the calling method!
 */
char* get_config(char file[], char config_string[])
{
	struct uci_context *ctx;
	struct uci_ptr p;
	struct uci_package *pkg;
	char option_name[strlen(file)+strlen(config_string)+1];
	char *a;
	
	strcpy(option_name, file);
	strcat(option_name, ".");
	strcat(option_name, config_string);
	a = strdup(option_name);
	
	/* allocate new uci context */
	ctx = uci_alloc_context ();
	if (!ctx)
		return NULL;
	
	/* override the default config storage for development purposes */
	uci_set_confdir(ctx, "./test/config");
	uci_load(ctx, file, &pkg);
	
	if (uci_lookup_ptr (ctx, &p, a, true) != UCI_OK)
	{
		return NULL;
	}
	
	uci_free_context(ctx);
	free (a);
	
	return strdup(p.o->v.string);
}

int file_read_line(char path[], char *buffer, int length)
{
	FILE *stream;
	if((stream = fopen(path,"r")) == NULL) {
		perror(path);
		return 1;
	}
	
	if ((fgets(buffer, length, stream)) != NULL)
		printf("The string is %s\n", buffer);
	else
		return 1;
	
	if (fclose(stream)) {
		perror("fclose error");
		return 1;
	}
	return 0;
}

/* ip test */
static struct
{
	int ifindex;
	int family;
	int oneline;
	int showqueue;
	inet_prefix pfx;
	int scope, scopemask;
	int flags, flagmask;
	int up;
	char *label;
	int flushed;
	char *flushb;
	int flushp;
	int flushe;
	int group;
} filter;

static int do_link;

struct nlmsg_list
{
	struct nlmsg_list *next;
	struct nlmsghdr	  h;
};

struct nlmsg_chain
{
	struct nlmsg_list *head;
	struct nlmsg_list *tail;
};

void ipaddr_reset_filter(int oneline)
{
	memset(&filter, 0, sizeof(filter));
	filter.oneline = oneline;
}

int preferred_family = AF_UNSPEC;
int show_stats = 0;
int show_details = 0;
int oneline = 0;
char * _SL_ = NULL;
struct rtnl_handle rth = { .fd = -1 };

static int store_nlmsg(const struct sockaddr_nl *who, struct nlmsghdr *n, void *arg)
{
	struct nlmsg_chain *lchain = (struct nlmsg_chain *)arg;
	struct nlmsg_list *h;

	h = malloc(n->nlmsg_len+sizeof(void*));
	if (h == NULL)
		return -1;

	memcpy(&h->h, n, n->nlmsg_len);
	h->next = NULL;

	if (lchain->tail)
		lchain->tail->next = h;
	else
		lchain->head = h;
	lchain->tail = h;

	ll_remember_index(who, n, NULL);
	return 0;
}

static void print_link_flags(FILE *fp, unsigned flags, unsigned mdown)
{
	fprintf(fp, "<");
	if (flags & IFF_UP && !(flags & IFF_RUNNING))
		fprintf(fp, "NO-CARRIER%s", flags ? "," : "");
	flags &= ~IFF_RUNNING;
#define _PF(f) if (flags&IFF_##f) { \
                  flags &= ~IFF_##f ; \
                  fprintf(fp, #f "%s", flags ? "," : ""); }
	_PF(LOOPBACK);
	_PF(BROADCAST);
	_PF(POINTOPOINT);
	_PF(MULTICAST);
	_PF(NOARP);
	_PF(ALLMULTI);
	_PF(PROMISC);
	_PF(MASTER);
	_PF(SLAVE);
	_PF(DEBUG);
	_PF(DYNAMIC);
	_PF(AUTOMEDIA);
	_PF(PORTSEL);
	_PF(NOTRAILERS);
	_PF(UP);
	_PF(LOWER_UP);
	_PF(DORMANT);
	_PF(ECHO);
#undef _PF
        if (flags)
		fprintf(fp, "%x", flags);
	if (mdown)
		fprintf(fp, ",M-DOWN");
	fprintf(fp, "> ");
}

static const char *oper_states[] = {
	"UNKNOWN", "NOTPRESENT", "DOWN", "LOWERLAYERDOWN", 
	"TESTING", "DORMANT",	 "UP"
};

static void print_operstate(FILE *f, __u8 state)
{
	if (state >= sizeof(oper_states)/sizeof(oper_states[0]))
		fprintf(f, "state %#x ", state);
	else
		fprintf(f, "state %s ", oper_states[state]);
}

static const char *link_modes[] = {
	"DEFAULT", "DORMANT"
};

static void print_linkmode(FILE *f, struct rtattr *tb)
{
	unsigned int mode = rta_getattr_u8(tb);

	if (mode >= sizeof(link_modes) / sizeof(link_modes[0]))
		fprintf(f, "mode %d ", mode);
	else
		fprintf(f, "mode %s ", link_modes[mode]);
}

static void print_queuelen(FILE *f, struct rtattr *tb[IFLA_MAX + 1])
{
	int qlen;

	if (tb[IFLA_TXQLEN])
		qlen = *(int *)RTA_DATA(tb[IFLA_TXQLEN]);
	else {
		struct ifreq ifr;
		int s = socket(AF_INET, SOCK_STREAM, 0);

		if (s < 0)
			return;

		memset(&ifr, 0, sizeof(ifr));
		strcpy(ifr.ifr_name, rta_getattr_str(tb[IFLA_IFNAME]));
		if (ioctl(s, SIOCGIFTXQLEN, &ifr) < 0) {
			fprintf(f, "ioctl(SIOCGIFTXQLEN) failed\n");
			close(s);
			return;
		}
		close(s);
		qlen = ifr.ifr_qlen;
	}
	if (qlen)
		fprintf(f, "qlen %d", qlen);
}


const char * ll_type_n2a(int type, char *buf, int len)
{
#define __PF(f,n) { ARPHRD_##f, #n },
static const struct {
	int type;
	const char *name;
} arphrd_names[] = {
{ 0, "generic" },
__PF(ETHER,ether)
__PF(EETHER,eether)
__PF(AX25,ax25)
__PF(PRONET,pronet)
__PF(CHAOS,chaos)
__PF(IEEE802,ieee802)
__PF(ARCNET,arcnet)
__PF(APPLETLK,atalk)
__PF(DLCI,dlci)
__PF(ATM,atm)
__PF(METRICOM,metricom)
__PF(IEEE1394,ieee1394)
__PF(INFINIBAND,infiniband)
__PF(SLIP,slip)
__PF(CSLIP,cslip)
__PF(SLIP6,slip6)
__PF(CSLIP6,cslip6)
__PF(RSRVD,rsrvd)
__PF(ADAPT,adapt)
__PF(ROSE,rose)
__PF(X25,x25)
__PF(HWX25,hwx25)
__PF(CAN,can)
__PF(PPP,ppp)
__PF(HDLC,hdlc)
__PF(LAPB,lapb)
__PF(DDCMP,ddcmp)
__PF(RAWHDLC,rawhdlc)
__PF(TUNNEL,ipip)
__PF(TUNNEL6,tunnel6)
__PF(FRAD,frad)
__PF(SKIP,skip)
__PF(LOOPBACK,loopback)
__PF(LOCALTLK,ltalk)
__PF(FDDI,fddi)
__PF(BIF,bif)
__PF(SIT,sit)
__PF(IPDDP,ip/ddp)
__PF(IPGRE,gre)
__PF(PIMREG,pimreg)
__PF(HIPPI,hippi)
__PF(ASH,ash)
__PF(ECONET,econet)
__PF(IRDA,irda)
__PF(FCPP,fcpp)
__PF(FCAL,fcal)
__PF(FCPL,fcpl)
__PF(FCFABRIC,fcfb0)
__PF(FCFABRIC+1,fcfb1)
__PF(FCFABRIC+2,fcfb2)
__PF(FCFABRIC+3,fcfb3)
__PF(FCFABRIC+4,fcfb4)
__PF(FCFABRIC+5,fcfb5)
__PF(FCFABRIC+6,fcfb6)
__PF(FCFABRIC+7,fcfb7)
__PF(FCFABRIC+8,fcfb8)
__PF(FCFABRIC+9,fcfb9)
__PF(FCFABRIC+10,fcfb10)
__PF(FCFABRIC+11,fcfb11)
__PF(FCFABRIC+12,fcfb12)
__PF(IEEE802_TR,tr)
__PF(IEEE80211,ieee802.11)
__PF(IEEE80211_PRISM,ieee802.11/prism)
__PF(IEEE80211_RADIOTAP,ieee802.11/radiotap)
__PF(IEEE802154, ieee802.15.4)
__PF(PHONET, phonet)
__PF(PHONET_PIPE, phonet_pipe)
__PF(CAIF, caif)

__PF(NONE, none)
__PF(VOID,void)
};
#undef __PF

        int i;
        for (i=0; i<sizeof(arphrd_names)/sizeof(arphrd_names[0]); i++) {
                 if (arphrd_names[i].type == type)
			return arphrd_names[i].name;
	}
        snprintf(buf, len, "[%d]", type);
        return buf;
}

const char *ll_addr_n2a(unsigned char *addr, int alen, int type, char *buf, int blen)
{
	int i;
	int l;

	if (alen == 4 &&
	    (type == ARPHRD_TUNNEL || type == ARPHRD_SIT || type == ARPHRD_IPGRE)) {
		return inet_ntop(AF_INET, addr, buf, blen);
	}
	if (alen == 16 && type == ARPHRD_TUNNEL6) {
		return inet_ntop(AF_INET6, addr, buf, blen);
	}
	l = 0;
	for (i=0; i<alen; i++) {
		if (i==0) {
			snprintf(buf+l, blen, "%02x", addr[i]);
			blen -= 2;
			l += 2;
		} else {
			snprintf(buf+l, blen, ":%02x", addr[i]);
			blen -= 3;
			l += 3;
		}
	}
	return buf;
}

int print_linkinfo(const struct sockaddr_nl *who,
		   struct nlmsghdr *n, void *arg)
{
	FILE *fp = (FILE*)arg;
	struct ifinfomsg *ifi = NLMSG_DATA(n);
	struct rtattr * tb[IFLA_MAX+1];
	int len = n->nlmsg_len;
	unsigned m_flag = 0;
	
	fprintf(fp, "\n start of print_linkinfo\n");

	if (n->nlmsg_type != RTM_NEWLINK && n->nlmsg_type != RTM_DELLINK)
		return 0;

	len -= NLMSG_LENGTH(sizeof(*ifi));
	if (len < 0)
		return -1;

	if (filter.ifindex && ifi->ifi_index != filter.ifindex)
		return 0;
	if (filter.up && !(ifi->ifi_flags&IFF_UP))
		return 0;

	parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), len);
	if (tb[IFLA_IFNAME] == NULL) {
		fprintf(stderr, "BUG: device with ifindex %d has nil ifname\n", ifi->ifi_index);
	}
	if (filter.label &&
	    (!filter.family || filter.family == AF_PACKET) &&
	    fnmatch(filter.label, RTA_DATA(tb[IFLA_IFNAME]), 0))
		return 0;

	if (tb[IFLA_GROUP]) {
		int group = *(int*)RTA_DATA(tb[IFLA_GROUP]);
		if (group != filter.group)
			return -1;
	}

	if (n->nlmsg_type == RTM_DELLINK)
		fprintf(fp, "Deleted ");

	fprintf(fp, "%d: %s", ifi->ifi_index,
		tb[IFLA_IFNAME] ? rta_getattr_str(tb[IFLA_IFNAME]) : "<nil>");

	if (tb[IFLA_LINK]) {
		SPRINT_BUF(b1);
		int iflink = *(int*)RTA_DATA(tb[IFLA_LINK]);
		if (iflink == 0)
			fprintf(fp, "@NONE: ");
		else {
			fprintf(fp, "@%s: ", ll_idx_n2a(iflink, b1));
			m_flag = ll_index_to_flags(iflink);
			m_flag = !(m_flag & IFF_UP);
		}
	} else {
		fprintf(fp, ": ");
	}
	print_link_flags(fp, ifi->ifi_flags, m_flag);
	

	if (tb[IFLA_MTU])
		fprintf(fp, "mtu %u ", *(int*)RTA_DATA(tb[IFLA_MTU]));
	if (tb[IFLA_QDISC])
		fprintf(fp, "qdisc %s ", rta_getattr_str(tb[IFLA_QDISC]));
	if (tb[IFLA_MASTER]) {
		SPRINT_BUF(b1);
		fprintf(fp, "master %s ", ll_idx_n2a(*(int*)RTA_DATA(tb[IFLA_MASTER]), b1));
	}

	if (tb[IFLA_OPERSTATE])
		print_operstate(fp, rta_getattr_u8(tb[IFLA_OPERSTATE]));

	if (do_link && tb[IFLA_LINKMODE])
		print_linkmode(fp, tb[IFLA_LINKMODE]);

	if (filter.showqueue)
		print_queuelen(fp, tb);

	if (!filter.family || filter.family == AF_PACKET) {
		SPRINT_BUF(b1);
		fprintf(fp, "%s", _SL_);
		fprintf(fp, "    link/%s ", ll_type_n2a(ifi->ifi_type, b1, sizeof(b1)));

		if (tb[IFLA_ADDRESS]) {
			fprintf(fp, "%s", ll_addr_n2a(RTA_DATA(tb[IFLA_ADDRESS]),
						      RTA_PAYLOAD(tb[IFLA_ADDRESS]),
						      ifi->ifi_type,
						      b1, sizeof(b1)));
		}
		if (tb[IFLA_BROADCAST]) {
			if (ifi->ifi_flags&IFF_POINTOPOINT)
				fprintf(fp, " peer ");
			else
				fprintf(fp, " brd ");
			fprintf(fp, "%s", ll_addr_n2a(RTA_DATA(tb[IFLA_BROADCAST]),
						      RTA_PAYLOAD(tb[IFLA_BROADCAST]),
						      ifi->ifi_type,
						      b1, sizeof(b1)));
		}
	}

/*	if (do_link && tb[IFLA_PROMISCUITY] && show_details)
		fprintf(fp, " promiscuity %u ",
			*(int*)RTA_DATA(tb[IFLA_PROMISCUITY]));

	if (do_link && tb[IFLA_LINKINFO] && show_details)
		print_linktype(fp, tb[IFLA_LINKINFO]);

	if (do_link && tb[IFLA_IFALIAS])
		fprintf(fp,"\n    alias %s", 
			rta_getattr_str(tb[IFLA_IFALIAS]));
/*
	if (do_link && show_stats) {
		if (tb[IFLA_STATS64])
			print_link_stats64(fp, RTA_DATA(tb[IFLA_STATS64]));
		else if (tb[IFLA_STATS])
			print_link_stats(fp, RTA_DATA(tb[IFLA_STATS]));
	}

	if (do_link && tb[IFLA_VFINFO_LIST] && tb[IFLA_NUM_VF]) {
		struct rtattr *i, *vflist = tb[IFLA_VFINFO_LIST];
		int rem = RTA_PAYLOAD(vflist);
		for (i = RTA_DATA(vflist); RTA_OK(i, rem); i = RTA_NEXT(i, rem))
			print_vfinfo(fp, i);
	}
*/
	fprintf(fp, "\n end of print_linkinfo");
	fprintf(fp, "\n");
	fflush(fp);
	return 0;
}

static int flush_update(void)
{
	if (rtnl_send_check(&rth, filter.flushb, filter.flushp) < 0) {
		perror("Failed to send flush request");
		return -1;
	}
	filter.flushp = 0;
	return 0;
}

#define DN_MAXADDL 20
struct dn_naddr
{
        unsigned short          a_len;
        unsigned char a_addr[DN_MAXADDL];
};

const char *rt_addr_n2a(int af, int len, const void *addr, char *buf, int buflen)
{
	switch (af) {
	case AF_INET:
	case AF_INET6:
		return inet_ntop(af, addr, buf, buflen);
/*	case AF_IPX:
		return ipx_ntop(af, addr, buf, buflen);
	case AF_DECnet:
	{
		struct dn_naddr dna = { 2, { 0, 0, }};
		memcpy(dna.a_addr, addr, 2);
		return dnet_ntop(af, &dna, buf, buflen);
	}*/
	default:
		return "???";
	}
}

static char * rtnl_rtscope_tab[256] = {
	"global",
};

static int rtnl_rtscope_init;

static void rtnl_rtscope_initialize(void)
{
	rtnl_rtscope_init = 1;
	rtnl_rtscope_tab[255] = "nowhere";
	rtnl_rtscope_tab[254] = "host";
	rtnl_rtscope_tab[253] = "link";
	rtnl_rtscope_tab[200] = "site";
/*	rtnl_tab_initialize(CONFDIR "/rt_scopes",
			    rtnl_rtscope_tab, 256);*/
}

const char *rtnl_rtscope_n2a(int id, char *buf, int len)
{
	if (id<0 || id>=256) {
		snprintf(buf, len, "%d", id);
		return buf;
	}
	if (!rtnl_rtscope_tab[id]) {
		if (!rtnl_rtscope_init)
			rtnl_rtscope_initialize();
	}
	if (rtnl_rtscope_tab[id])
		return rtnl_rtscope_tab[id];
	snprintf(buf, len, "%d", id);
	return buf;
}

int inet_addr_match(const inet_prefix *a, const inet_prefix *b, int bits)
{
	const __u32 *a1 = a->data;
	const __u32 *a2 = b->data;
	int words = bits >> 0x05;

	bits &= 0x1f;

	if (words)
		if (memcmp(a1, a2, words << 2))
			return -1;

	if (bits) {
		__u32 w1, w2;
		__u32 mask;

		w1 = a1[words];
		w2 = a2[words];

		mask = htonl((0xffffffff) << (0x20 - bits));

		if ((w1 ^ w2) & mask)
			return 1;
	}

	return 0;
}

int print_addrinfo(const struct sockaddr_nl *who, struct nlmsghdr *n,
		   void *arg)
{
	FILE *fp = (FILE*)arg;
	struct ifaddrmsg *ifa = NLMSG_DATA(n);
	int len = n->nlmsg_len;
	struct rtattr * rta_tb[IFA_MAX+1];
	char abuf[256];
	SPRINT_BUF(b1);

	if (n->nlmsg_type != RTM_NEWADDR && n->nlmsg_type != RTM_DELADDR)
		return 0;
	len -= NLMSG_LENGTH(sizeof(*ifa));
	if (len < 0) {
		fprintf(stderr, "BUG: wrong nlmsg len %d\n", len);
		return -1;
	}

	if (filter.flushb && n->nlmsg_type != RTM_NEWADDR)
		return 0;

	parse_rtattr(rta_tb, IFA_MAX, IFA_RTA(ifa), n->nlmsg_len - NLMSG_LENGTH(sizeof(*ifa)));

	if (!rta_tb[IFA_LOCAL])
		rta_tb[IFA_LOCAL] = rta_tb[IFA_ADDRESS];
	if (!rta_tb[IFA_ADDRESS])
		rta_tb[IFA_ADDRESS] = rta_tb[IFA_LOCAL];

	if (filter.ifindex && filter.ifindex != ifa->ifa_index)
		return 0;
	if ((filter.scope^ifa->ifa_scope)&filter.scopemask)
		return 0;
	if ((filter.flags^ifa->ifa_flags)&filter.flagmask)
		return 0;
	if (filter.label) {
		SPRINT_BUF(b1);
		const char *label;
		if (rta_tb[IFA_LABEL])
			label = RTA_DATA(rta_tb[IFA_LABEL]);
		else
			label = ll_idx_n2a(ifa->ifa_index, b1);
		if (fnmatch(filter.label, label, 0) != 0)
			return 0;
	}
	if (filter.pfx.family) {
		if (rta_tb[IFA_LOCAL]) {
			inet_prefix dst;
			memset(&dst, 0, sizeof(dst));
			dst.family = ifa->ifa_family;
			memcpy(&dst.data, RTA_DATA(rta_tb[IFA_LOCAL]), RTA_PAYLOAD(rta_tb[IFA_LOCAL]));
			if (inet_addr_match(&dst, &filter.pfx, filter.pfx.bitlen))
				return 0;
		}
	}

	if (filter.family && filter.family != ifa->ifa_family)
		return 0;

	if (filter.flushb) {
		struct nlmsghdr *fn;
		if (NLMSG_ALIGN(filter.flushp) + n->nlmsg_len > filter.flushe) {
			if (flush_update())
				return -1;
		}
		fn = (struct nlmsghdr*)(filter.flushb + NLMSG_ALIGN(filter.flushp));
		memcpy(fn, n, n->nlmsg_len);
		fn->nlmsg_type = RTM_DELADDR;
		fn->nlmsg_flags = NLM_F_REQUEST;
		fn->nlmsg_seq = ++rth.seq;
		filter.flushp = (((char*)fn) + n->nlmsg_len) - filter.flushb;
		filter.flushed++;
		if (show_stats < 2)
			return 0;
	}

	if (n->nlmsg_type == RTM_DELADDR)
		fprintf(fp, "Deleted ");

	if (filter.oneline || filter.flushb)
		fprintf(fp, "%u: %s", ifa->ifa_index, ll_index_to_name(ifa->ifa_index));
	if (ifa->ifa_family == AF_INET)
		fprintf(fp, "    inet ");
	else if (ifa->ifa_family == AF_INET6)
		fprintf(fp, "    inet6 ");
	else if (ifa->ifa_family == AF_DECnet)
		fprintf(fp, "    dnet ");
	else if (ifa->ifa_family == AF_IPX)
		fprintf(fp, "     ipx ");
	else
		fprintf(fp, "    family %d ", ifa->ifa_family);

	if (rta_tb[IFA_LOCAL]) {
		fprintf(fp, "%s", rt_addr_n2a(ifa->ifa_family,
					      RTA_PAYLOAD(rta_tb[IFA_LOCAL]),
					      RTA_DATA(rta_tb[IFA_LOCAL]),
					      abuf, sizeof(abuf)));

		if (rta_tb[IFA_ADDRESS] == NULL ||
		    memcmp(RTA_DATA(rta_tb[IFA_ADDRESS]), RTA_DATA(rta_tb[IFA_LOCAL]), 4) == 0) {
			fprintf(fp, "/%d ", ifa->ifa_prefixlen);
		} else {
			fprintf(fp, " peer %s/%d ",
				rt_addr_n2a(ifa->ifa_family,
					    RTA_PAYLOAD(rta_tb[IFA_ADDRESS]),
					    RTA_DATA(rta_tb[IFA_ADDRESS]),
					    abuf, sizeof(abuf)),
				ifa->ifa_prefixlen);
		}
	}

	if (rta_tb[IFA_BROADCAST]) {
		fprintf(fp, "brd %s ",
			rt_addr_n2a(ifa->ifa_family,
				    RTA_PAYLOAD(rta_tb[IFA_BROADCAST]),
				    RTA_DATA(rta_tb[IFA_BROADCAST]),
				    abuf, sizeof(abuf)));
	}
	if (rta_tb[IFA_ANYCAST]) {
		fprintf(fp, "any %s ",
			rt_addr_n2a(ifa->ifa_family,
				    RTA_PAYLOAD(rta_tb[IFA_ANYCAST]),
				    RTA_DATA(rta_tb[IFA_ANYCAST]),
				    abuf, sizeof(abuf)));
	}
	fprintf(fp, "scope %s ", rtnl_rtscope_n2a(ifa->ifa_scope, b1, sizeof(b1)));

	fprintf(fp, "\n");
	fflush(fp);
	return 0;
}

static int print_selected_addrinfo(int ifindex, struct nlmsg_list *ainfo, FILE *fp)
{
	for ( ;ainfo ;  ainfo = ainfo->next) {
		struct nlmsghdr *n = &ainfo->h;
		struct ifaddrmsg *ifa = NLMSG_DATA(n);

		if (n->nlmsg_type != RTM_NEWADDR)
			continue;

		if (n->nlmsg_len < NLMSG_LENGTH(sizeof(ifa)))
			return -1;

		if (ifa->ifa_index != ifindex ||
		    (filter.family && filter.family != ifa->ifa_family))
			continue;

		print_addrinfo(NULL, n, fp);
	}
	return 0;
}

static void ipaddr_filter(struct nlmsg_chain *linfo, struct nlmsg_chain *ainfo)
{
	struct nlmsg_list *l, **lp;

	lp = &linfo->head;
	while ( (l = *lp) != NULL) {
		int ok = 0;
		int missing_net_address = 1;
		struct ifinfomsg *ifi = NLMSG_DATA(&l->h);
		struct nlmsg_list *a;

		for (a = ainfo->head; a; a = a->next) {
			struct nlmsghdr *n = &a->h;
			struct ifaddrmsg *ifa = NLMSG_DATA(n);

			if (ifa->ifa_index != ifi->ifi_index)
				continue;
			missing_net_address = 0;
			if (filter.family && filter.family != ifa->ifa_family)
				continue;
			if ((filter.scope^ifa->ifa_scope)&filter.scopemask)
				continue;
			if ((filter.flags^ifa->ifa_flags)&filter.flagmask)
				continue;
			if (filter.pfx.family || filter.label) {
				struct rtattr *tb[IFA_MAX+1];
				parse_rtattr(tb, IFA_MAX, IFA_RTA(ifa), IFA_PAYLOAD(n));
				if (!tb[IFA_LOCAL])
					tb[IFA_LOCAL] = tb[IFA_ADDRESS];

				if (filter.pfx.family && tb[IFA_LOCAL]) {
					inet_prefix dst;
					memset(&dst, 0, sizeof(dst));
					dst.family = ifa->ifa_family;
					memcpy(&dst.data, RTA_DATA(tb[IFA_LOCAL]), RTA_PAYLOAD(tb[IFA_LOCAL]));
					if (inet_addr_match(&dst, &filter.pfx, filter.pfx.bitlen))
						continue;
				}
				if (filter.label) {
					SPRINT_BUF(b1);
					const char *label;
					if (tb[IFA_LABEL])
						label = RTA_DATA(tb[IFA_LABEL]);
					else
						label = ll_idx_n2a(ifa->ifa_index, b1);
					if (fnmatch(filter.label, label, 0) != 0)
						continue;
				}
			}

			ok = 1;
			break;
		}
		if (missing_net_address &&
		    (filter.family == AF_UNSPEC || filter.family == AF_PACKET))
			ok = 1;
		if (!ok) {
			*lp = l->next;
			free(l);
		} else
			lp = &l->next;
	}
}

int main(int argc, char **argv)
{
	/* storage for the nodewatcher configuration */
	char *error_level = get_config("nodewatcher", "@nodewatcher[0].error_level");
	char *logfile = get_config("nodewatcher", "@nodewatcher[0].logfile");
	char *data_file = get_config("nodewatcher", "@nodewatcher[0].data_file");
	char *mesh_interface = get_config("nodewatcher", "@network[0].mesh_interface");
	char *client_interfaces = get_config("nodewatcher", "@network[0].client_interfaces");
	
	/* version of the program */
	char version[] = "0";
	/* hostname of the system */
	char hostname[128];
	/* batman advanced version */
	char batman_adv_version[4096];	
	
	/* structs to store systeminformations parsed by some methods */
	struct sysinfo s_sysinfo;
	struct cpuinfo s_cpuinfo;
	struct utsname kernelinfo;
	struct openwrtinfo s_openwrtinfo;
	struct firmwareinfo s_firmwareinfo;
	
	/* structs for the interfaces */
	char          buf[1024];
	struct ifconf ifc;
	struct ifreq *ifr;
	int           sck;
	int           nInterfaces;
	int           i;
	char mac[SIZEOFMAC];
	
	
	/* ip test */
	struct nlmsg_chain linfo = { NULL, NULL};
	struct nlmsg_chain ainfo = { NULL, NULL};
	struct nlmsg_list *l;
	int no_link = 0;
	
	_SL_ = oneline ? "\\" : "\n" ;
	
	if (rtnl_open(&rth, 0) < 0)
		exit(1);
	
	ipaddr_reset_filter(oneline);
	filter.showqueue = 1;
	
	if (rtnl_wilddump_request(&rth, preferred_family, RTM_GETLINK) < 0) {
		perror("Cannot send dump request");
		exit(1);
	}
	
	if (rtnl_dump_filter(&rth, store_nlmsg, &linfo) < 0) {
		fprintf(stderr, "Dump terminated\n");
		exit(1);
	}
	
	if (filter.family != AF_PACKET) {
		if (filter.oneline)
			no_link = 1;

		if (rtnl_wilddump_request(&rth, filter.family, RTM_GETADDR) < 0) {
			perror("Cannot send dump request");
			exit(1);
		}

		if (rtnl_dump_filter(&rth, store_nlmsg, &ainfo) < 0) {
			fprintf(stderr, "Dump terminated\n");
			exit(1);
		}

		ipaddr_filter(&linfo, &ainfo);
	}
	
	for (l = linfo.head; l; l = l->next) {
		if (no_link || print_linkinfo(NULL, &l->h, stdout) == 0) {
			struct ifinfomsg *ifi = NLMSG_DATA(&l->h);
			if (filter.family != AF_PACKET) {
				print_selected_addrinfo(ifi->ifi_index,
							ainfo.head, stdout);
			}
		}
	}
	
	rtnl_close(&rth);
	
	/* ifconfig */
/*	struct ifaddrs *ifap, *ifa;
	struct ifreq ifr;
	const struct sockaddr *sdl = NULL;
	prop_dictionary_t env, oenv;
	int idx;
	char *p;*/
	
	/* get the hostname */
	if(gethostname(hostname, sizeof hostname) != 0)
		strcpy(hostname, "");
	
	/* get batman advanced version */
	if(file_read_line("/sys/module/batman_adv/version", batman_adv_version, sizeof batman_adv_version) != 0)
		strcpy(batman_adv_version, "");
	
	/* get system information */
	if(sysinfo(&s_sysinfo) != 0)
		printf("Error fetching sysinfo");
	
	/* parse /proc/cpuinfo */
	if(get_cpuinfo(&s_cpuinfo) != 0)
		printf("Error fetching cpuinfo");
	
	if(uname(&kernelinfo) != 0)
		printf("Error fetching kernelinfo");
	
	if(get_openwrtinfo(&s_openwrtinfo) != 0)
		printf("Error fetching openwrtinfo");

	if(get_firmwareinfo(&s_firmwareinfo) != 0)
		printf("Error fetching openwrtinfo");
	
	
/*	printf("error_level: %s\n", error_level);
	printf("logfile: %s\n", logfile);
	printf("data_file: %s\n", data_file);*/
	
	/* print all values into an xml */
	printf("<hostname>%s<hostname>", hostname);	
	printf("<uptime>%ld</uptime>", s_sysinfo.uptime);
	printf("<memory_total>%lu</memory_total>", s_sysinfo.totalram);
/*	dont know how to get this without manual parsing
	printf("<memory_caching>%lu</memory_caching>", memory_caching);*/
	printf("<memory_buffering>%lu</memory_buffering>", s_sysinfo.bufferram);
	printf("<memory_free>%lu</memory_free>", s_sysinfo.freeram);
	printf("<loadavg_1>%lu</loadavg_1>", s_sysinfo.loads[0]);
	printf("<loadavg_5>%lu</loadavg_5>", s_sysinfo.loads[1]);
	printf("<loadavg_15>%lu</loadavg_15>", s_sysinfo.loads[2]);
	printf("<cpu>%s</cpu>", s_cpuinfo.cpu);
	printf("<chipset>%s</chipset>", s_cpuinfo.chipset);
	printf("<local_time>%ld</local_time>", (long)time(NULL));
	printf("<nodewatcher_version>%s</nodewatcher_version>", version);
	printf("<kernel_version>%s</kernel_version>", kernelinfo.release);
	printf("<batman_adv_version>%s</batman_adv_version>", batman_adv_version);
	printf("<distversion>%s</distversion>", s_openwrtinfo.distrib_release);
	printf("<distname>%s</distname>", s_openwrtinfo.distrib_id);
	printf("<firmware_version>%s</firmware_version>", s_firmwareinfo.firmware_version);
	printf("<firmware_revision>%s</firmware_revision>", s_firmwareinfo.firmware_revision);
	printf("<openwrt_core_revision>%s</openwrt_core_revision>", s_firmwareinfo.openwrt_core_revision);
	printf("<openwrt_feeds_packages_revision>%s</openwrt_feeds_packages_revision>", s_firmwareinfo.openwrt_feeds_packages_revision);
	printf("\n");

	/* Get a socket handle. */
	sck = socket(AF_INET, SOCK_DGRAM, 0);
	if(sck < 0)
	{
		perror("socket");
		return 1;
	}
	
	/* Query available interfaces. */
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if(ioctl(sck, SIOCGIFCONF, &ifc) < 0)
	{
		perror("ioctl(SIOCGIFCONF)");
		return 1;
	}
	
	/* Iterate through the list of interfaces. */
	ifr         = ifc.ifc_req;
	nInterfaces = ifc.ifc_len / sizeof(struct ifreq);
	for(i = 0; i < nInterfaces; i++)
	{
		struct ifreq *item = &ifr[i];
		
		/* Show the device name */
		printf("%s:", item->ifr_name);
		
		if(ioctl(sck, SIOCGIFNAME, item) >= 0)
			printf(", NAME %s", inet_ntoa(((struct sockaddr_in *)&item->ifr_broadaddr)->sin_addr));
		
		
		/* show IPv4 address */
		printf(", IP %s", inet_ntoa(((struct sockaddr_in *)&item->ifr_addr)->sin_addr));
		
		
		/* Get the MAC address */
		if(ioctl(sck, SIOCGIFHWADDR, item) >= 0) {
			snprintf(mac, SIZEOFMAC*sizeof(char), "%02x:%02x:%02x:%02x:%02x:%02x",
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[0],
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[1],
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[2],
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[3],
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[4],
				(int)((unsigned char *)item->ifr_hwaddr.sa_data)[5]);
			printf(", %s", mac);
		}
	
		/* Get the broadcast address (added by Eric) */
		if(ioctl(sck, SIOCGIFBRDADDR, item) >= 0)
			printf(", BROADCAST %s", inet_ntoa(((struct sockaddr_in *)&item->ifr_broadaddr)->sin_addr));
		
		printf("\n");
	}

	
	
	/* free allocated memory */
	free(logfile);
	free(error_level);
	free(data_file);
	free(mesh_interface);
	free(client_interfaces);
		
	return 0;
}

#endif